require 'webrick' 
require 'xmlrpc/server'
require 'html-formatting'
require 'paths'
require 'pp'
require 'app'

include WEBrick
include HtmlFormatting

class AppWebInterface

  def initialize(app)
    @server = HTTPServer.new(:Port => 8080)
    @app = app
    @dispatched = []
    @current_page_name = :no_page_yet
    install_UI_servlets
    install_testing_servlets
    install_utility_servlets
    trap("INT"){ @server.shutdown } 
    @server.start
  end

  def install_generic_proc(mount_point, &block)
    @server.mount_proc(mount_point,
                       proc { | request, response |
                         response.body = block.call(request, @app)
                         raise HTTPStatus::OK
                       })
  end

  def install_command(command, *required_args)
    install_generic_proc("/#{command}") { | request, app |
      @current_page_name = dispatch(command,
                                     values(request, *required_args))
      render(@current_page_name)
    }
  end

  def dispatch(command, args)
    @dispatched << [command, args]
    @app.send(command, *args)
  end

  def values(request, *args)
    args.collect { | arg |
      request.query[arg]
    }
  end

  def render(page_name)
    @current_xhtml = send("#{page_name}_for", @app)
  end



  def install_UI_servlets
    install_generic_proc('/') { | request, app | 
      beginning_page_for(app)
    }
    install_command(:login, 'login', 'password')
    install_command(:new_case)
    install_command(:record_case, 'client', 'clinic_id')
    install_command(:add_visit)
    install_command(:add_audit)
    install_command(:record_visit, 'diagnosis', 'charges')
    install_command(:record_audit, 'auditor', 'variance')
  end

  def install_utility_servlets
    install_generic_proc('/ping') { | request, app |
      ping_page
    }

    install_generic_proc('/refresh') { | request, app |
      render(@current_page_name)
    }
  end


  def install_testing_servlets
    dumper_servlet = XMLRPC::WEBrickServlet.new
    @server.mount("/side-channel", dumper_servlet)

    dumper_servlet.add_handler("snapshot") {
      retval = '' 
      PP.pp(@dispatched, retval)
      retval
    }

    dumper_servlet.add_handler("replay") { | command_string |
      command_descriptions = eval(command_string)
      command_descriptions.each { | one |
        @current_page_name = dispatch(*one)
      }
      'return value irrelevant but it cannot be nil'
    }

  end


  # Page Generators

  def ping_page
    page("Server Alive")
  end

  def beginning_page_for(app)
    page('Welcome to the Case Management System',
         p({:align => 'center'},
           "Authorized Users", b("Only")),
         command_form('login',
                      labeled_textfield('login'),
                      labeled_textfield('password'),
                      submit('Login')))
  end


  def main_page_for(app)
    page("Cases Available to #{app.user.capitalize}",
         p({ :align => 'center'}, "No cases are available."),
         command_form('new_case',
                      submit('Add a New Case')))
  end


  def gather_case_input_page_for(app)
    page("Add a Case",
         command_form('record_case',
                      labeled_textfield('client'),
                      labeled_textfield('clinic_id'),
                      submit('Add')))
  end



  def visit_list(visits)
    unless visits.empty?
      p("Visits: ", br,
        ul(visits.collect { | visit |
             li("#{visit.diagnosis} (#{visit.charges})")
           }))
    end
  end

  def audit_list(audits)
    unless audits.empty?
      p("Audits: ", br,
        ul(audits.collect { | audit |
             li("#{audit.auditor} found #{audit.variance} variance.")
           }))
    end
  end



  def case_display_page_for(app)
    case_record = app.current_record
    page("Case #{case_record.clinic_id}",
         p("Owner: #{case_record.client}"),
         visit_list(case_record.visits),
         audit_list(case_record.audits),
         p(command_form('add_visit',
                        submit('Add a Visit'))),
         p(command_form('add_audit',
                        submit('Add an Audit Record'))))
  end


  def gather_visit_input_page_for(app)
    case_record = app.current_record
    page("Add a visit to case #{case_record.clinic_id}.",
         command_form('record_visit',
                      labeled_textfield('diagnosis'),
                      labeled_textfield('charges'),
                      submit('Add')))
  end

  def gather_audit_input_page_for(app)
    case_record = app.current_record
    page("Add an audit to case #{case_record.clinic_id}.",
         command_form('record_audit',
                      labeled_textfield('auditor'),
                      labeled_textfield('variance'),
                      submit('Add')))
  end
end


if __FILE__ == $0
  AppWebInterface.new(App.new)
end
